@@ -10,16 +10,19 @@
 # Author: Fabien BATTELLO <fabien.battello@orange.com> et al.
 
 import logging
-from dataclasses import dataclass
-from typing import Optional, overload
-
 import requests
+from dataclasses import dataclass
+from typing import Optional, Tuple, Generator, List, Union, overload
+from math import ceil
 
+from ..utils import is_interactive
 from ..model.entity import Entity
 from .constants import *
 from .entities import Entities
 from .batch import BatchOp
 from .types import Types
+from .contexts import Contexts
+from .subscriptions import Subscriptions
 from .exceptions import *
 
 logger = logging.getLogger(__name__)
@@ -44,7 +47,7 @@ class Client:
     As for now it focuses on the /entities/{entityId} endpoint.
     Allowed operations are :
     - create(), update(), upsert()
-    - retrieve(), exists()
+    - get(), exists()
     - delete()
 
     Update() and upsert() operations are not atomic, as they
@@ -140,13 +143,14 @@ class Client:
         self._entities = Entities(self, f"{self.url}/{ENDPOINT_ENTITIES}")
         self._batch = BatchOp(self, f"{self.url}/{ENDPOINT_BATCH}")
         self._types = Types(self, f"{self.url}/{ENDPOINT_TYPES}")
-        self._subscriptions = (
-            None  # TODO : create Subscriptions class and implement subscription stuff
-        )
+        self._contexts = Contexts(self, f"{self.url}/{ENDPOINT_CONTEXTS}")
+        self._subscriptions = Subscriptions(self, f"{self.url}/{ENDPOINT_SUBSCRIPTIONS}")
+
+        self.broker = Broker(Vendor.UNKNOWN, "N/A")
 
         # get status and retrieve Context Broker information
         status = self.is_connected(raise_for_disconnected=True)
-        if status:
+        if status and is_interactive():
             self.broker = Broker(*self.guess_vendor())
             print(self._welcome_message())
         else:
@@ -195,10 +199,10 @@ class Client:
             )
             r.raise_for_status()
         except Exception as e:
+            if is_interactive():
+                return False
             if raise_for_disconnected:
-                raise NgsiNotConnectedError(
-                    f"Cannot connect to Context Broker at {self.hostname}:{self.port}: {e}"
-                )
+                raise NgsiNotConnectedError(f"Cannot connect to Context Broker at {self.hostname}:{self.port}: {e}")
             else:
                 logger.error(e)
                 return False
@@ -220,6 +224,10 @@ class Client:
     def types(self):
         return self._types
 
+    @property
+    def contexts(self):
+        return self._contexts
+
     @property
     def subscriptions(self):
         return self._subscriptions
@@ -232,9 +240,7 @@ class Client:
         self.session.close()
 
     @overload
-    def create(
-        self, entity: Entity, skip: bool = False, overwrite: bool = False
-    ) -> Entity:
+    def create(self, entity: Entity, skip: bool = False, overwrite: bool = False) -> Entity:
         """Create an entity.
 
         Facade method for Entities.create().
@@ -256,9 +262,7 @@ class Client:
         ...
 
     @overload
-    def create(
-        self, entities: List[Entity], skip: bool = False, overwrite: bool = False
-    ):
+    def create(self, entities: List[Entity], skip: bool = False, overwrite: bool = False):
         """Create a batch of entities.
 
         Facade method for Batch.create().
@@ -291,8 +295,12 @@ class Client:
         else:
             return self.batch.create(_entities, skip, overwrite)
 
-    def retrieve(
-        self, eid: Union[EntityId, Entity], asdict: bool = False, **kwargs
+    def get(
+        self,
+        eid: Union[EntityId, Entity],
+        ctx: str = None,
+        asdict: bool = False,
+        **kwargs,
     ) -> Entity:
         """Retrieve an entity given its id.
 
@@ -303,6 +311,8 @@ class Client:
         ----------
         eid : Union[EntityId, Entity]
             The entity identifier or the entity instance
+        ctx : str
+            The context
         asdict : bool, optional
             If set (instead of returning an Entity) returns the raw API response (a Python dict that represents the JSON response), by default False
 
@@ -311,7 +321,7 @@ class Client:
         Entity
             The retrieved entity
         """
-        return self.entities.retrieve(eid)
+        return self.entities.get(eid, ctx, asdict, **kwargs)
 
     @overload
     def delete(self, eid: Union[EntityId, Entity]) -> bool:
@@ -351,15 +361,24 @@ class Client:
         """
         ...
 
-    def delete(
-        self, eids: Union[Union[EntityId, Entity], List[Union[EntityId, Entity]]]
-    ) -> bool:
+    def delete(self, eids: Union[Union[EntityId, Entity], List[Union[EntityId, Entity]]]) -> bool:
         if isinstance(eids, list):
             return self.batch.delete(eids)
         else:
             eid = eids
             return self.entities.delete(eid)
 
+    def delete_from_file(self, filename: str) -> Union[Entity, dict]:
+        """Delete in the broker all entities present in the JSON file.
+
+        Parameters
+        ----------
+        filename : str
+            Points to the JSON input file that contains entities.
+        """
+        entities = Entity.load(filename)
+        return self.delete(entities)
+
     def exists(self, eid: Union[EntityId, Entity]) -> bool:
         """Tests if an entity exists.
 
@@ -421,6 +440,17 @@ class Client:
         else:
             return self.batch.upsert(entities)
 
+    def bulk_import(self, filename: str) -> Union[Entity, dict]:
+        """Upsert all entities from a JSON file.
+
+        Parameters
+        ----------
+        filename : str
+            Points to the JSON input file that contains entities.
+        """
+        entities = Entity.load(filename)
+        return self.upsert(entities)
+
     @overload
     def update(self, entity: Entity) -> Optional[Entity]:
         """Update the entity.
@@ -457,19 +487,19 @@ class Client:
         """
         ...
 
-    def update(
-        self, entities: Union[Entity, List[Entity]]
-    ) -> Union[Optional[Entity], dict]:
+    def update(self, entities: Union[Entity, List[Entity]]) -> Union[Optional[Entity], dict]:
         if isinstance(entities, Entity):
             entity = entities
             return self.entities.update(entity)
         else:
             return self.batch.update(entities)
 
-    def query(self, type: str = None, q: str = None, **kwargs) -> List[Entity]:
+    def query_head(self, type: str = None, q: str = None, ctx: str = None, n: int = 5, **kwargs) -> List[Entity]:
         """Retrieve entities given its type and/or query string.
 
-        Facade method for Entities.query().
+        Retrieve up to PAGINATION_LIMIT_MAX entities.
+        Use query_all() to retrieve all entities.
+        Use entities.query() to deal with limit and offset on your own.
 
         Parameters
         ----------
@@ -477,7 +507,10 @@ class Client:
             The entity's type
         q: str
             The query string (NGSI-LD Query Language)
-
+        ctx: str
+            The context
+        n: int
+            The first n entities to be retrieved
         Returns
         -------
         list[Entity]
@@ -491,9 +524,72 @@ class Client:
         >>> with Client() as client:
         >>>     client.query(type="AgriFarm", q='contactPoint[email]=="wheatfarm@email.com"') # match type and query
         """
-        return self.entities.query(type, q)
+        return self.entities.query(type, q, ctx, limit=n)
 
-    def count(self, type: str = None, query: str = None, **kwargs) -> int:
+    def query_all(
+        self,
+        type: str = None,
+        q: str = None,
+        ctx: str = None,
+        limit: int = PAGINATION_LIMIT_MAX,
+        max: int = 1_000_000,
+        **kwargs,
+    ) -> List[Entity]:
+        """Retrieve entities given its type and/or query string.
+
+        Retrieve all entities by sending as many requests as needed, using pagination.
+        Assume data hold in memory. Should not be an issue except for very large datasets.
+
+        Parameters
+        ----------
+        etype : str
+            The entity's type
+        q: str
+            The query string (NGSI-LD Query Language)
+        ctx: str
+            The context
+        limit: int
+            The number of entities retrieved in each request
+
+        Returns
+        -------
+        list[Entity]
+            Retrieved entities matching the given type and/or query string
+
+        Example:
+        --------
+        >>> with Client() as client:
+        >>>     client.query_all(type="AgriFarm") # match a given type
+
+        >>> with Client() as client:
+        >>>     client.query_all(type="AgriFarm", q='contactPoint[email]=="wheatfarm@email.com"') # match type and query
+        """
+
+        entities: list[Entity] = []
+        count = self.entities.count(type, q)
+        if count > max:
+            raise NgsiClientTooManyResultsError(f"{count} results exceed maximum {max}")
+        for page in range(ceil(count / limit)):
+            entities.extend(self.entities.query(type, q, ctx, limit, page * limit))
+        return entities
+
+    def query_generator(
+        self,
+        type: str = None,
+        q: str = None,
+        ctx: str = None,
+        limit: int = PAGINATION_LIMIT_MAX,
+        batch: bool = False,
+        **kwargs,
+    ) -> Generator[Entity, None, None]:
+        count = self.entities.count(type, q)
+        for page in range(ceil(count / limit)):
+            if batch:
+                yield self.entities.query(type, q, ctx, limit, page * limit)
+            else:
+                yield from self.entities.query(type, q, ctx, limit, page * limit)
+
+    def count(self, type: str = None, q: str = None, **kwargs) -> int:
         """Return number of entities matching type and/or query string.
 
         Facade method for Entities.count().
@@ -518,11 +614,9 @@ class Client:
         >>> with Client() as client:
         >>>     client.count(type="AgriFarm", query='contactPoint[email]=="wheatfarm@email.com"') # match type and query
         """
-        return self.entities.count(type, query)
+        return self.entities.count(type, q)
 
-    def delete_where(
-        self, type: str = None, q: str = None, **kwargs
-    ) -> tuple[bool, dict]:
+    def delete_where(self, type: str = None, q: str = None, **kwargs):
         """Batch delete entities matching type and/or query string.
 
         Parameters
@@ -537,8 +631,9 @@ class Client:
         >>> with Client() as client:
         >>>     client.delete_where(type="AgriFarm", query='contactPoint[email]=="wheatfarm@email.com"') # match type and query
         """
-        entities = self.query(type, q, **kwargs)
-        return self.batch.delete(entities)
+        g = self.query_generator(type, q, batch=True, **kwargs)
+        for batch in g:
+            self.batch.delete(batch)
 
     def drop(self, type: str) -> None:
         """Batch delete entities matching the given type.
@@ -555,6 +650,9 @@ class Client:
         """
         self.delete_where(type=type)
 
+    def list_types(self) -> Optional[dict]:
+        return self.types.list()
+
     def purge(self) -> None:
         """Batch delete all entities.
 
@@ -563,9 +661,20 @@ class Client:
         >>> with Client() as client:
         >>>     client.purge()
         """
-        for type in self.types.available():
+        for type in self.types.list():
             self.drop(type)
 
+    def flush_all(self) -> None:
+        """Batch delete all entities and remove all contexts.
+
+        Example:
+        --------
+        >>> with Client() as client:
+        >>>     client.purge()
+        """
+        self.purge(type)
+        self.contexts.cleanup()
+
     def guess_vendor(self) -> tuple[Vendor, Version]:
         """Try to guess the Context Broker vendor.
 
@@ -582,12 +691,10 @@ class Client:
         >>>     print(client.guess_vendor())
         (<Vendor.ORIONLD: 'Orion-LD'>, 'post-v0.8.1')
         """
-        if version := self._broker_version_orionld():
-            return Vendor.ORIONLD, version
-        if version := self._broker_version_scorpio():
-            return Vendor.SCORPIO, version
-        if version := self._broker_version_stellio():
-            return Vendor.STELLIO, version
+        if broker := self._broker_version_orionld():
+            return broker
+        if broker := self._broker_version_java_spring():
+            return broker
         return Vendor.UNKNOWN, "N/A"
 
     def _broker_version_orionld(self) -> Optional[str]:
@@ -608,71 +715,57 @@ class Client:
         try:
             r = self.session.get(url, headers=headers)
             r.raise_for_status()
-            return r.json()["orionld version"]
+            return Vendor.ORIONLD, r.json()["orionld version"]
         except Exception:
             return None
 
-    def _broker_version_scorpio(self) -> Optional[str]:
-        """Requests the broker looking for Scorpio version.
+    def _broker_version_java_spring(self) -> Optional[Tuple[Vendor, str]]:
+        """Requests the Java-Spring based broker looking for Vendor and Version.
 
-        Targets the /health endpoint.
-        No /version endpoint as for now.
+        Targets the /actuator/info endpoint.
 
         Returns
         -------
-        Optional[str]
-            "N/A" if the broker is Scorpio else None
+        Optional[Tuple[Vendor, str]]
+            A tuple composed of the Vendor and the broker version
         """
-        url = f"{self.url}/scorpio/v1/info/health"
+        url = f"{self.url}/actuator"
         headers = {
             "Accept": "application/json",
             "Content-Type": None,
         }  # overrides session headers
         try:
-            r = self.session.get(url, headers=headers)
+            r = self.session.get(f"{url}/health", headers=headers)
             r.raise_for_status()
-            return "N/A"
         except Exception:
             return None
-
-    def _broker_version_stellio(self) -> Optional[str]:
-        """Requests the broker looking for Stellio version.
-
-        Targets the /actuator/health endpoint.
-        No /version endpoint as for now.
-
-        Returns
-        -------
-        Optional[str]
-            "N/A" if the broker is Stellio else None
-        """
-        url = f"{self.url}/actuator/health"
-        headers = {
-            "Accept": "application/json",
-            "Content-Type": None,
-        }  # overrides session headers
         try:
-            r = self.session.get(url, headers=headers)
+            r = self.session.get(f"{url}/info", headers=headers)
             r.raise_for_status()
-            return "N/A"
+            build = r.json()["build"]
+            version = build["version"]
+            group = build["group"]
+            if group == "eu.neclab.ngsildbroker":
+                vendor = Vendor.SCORPIO
+            elif group == "com.egm.stellio":
+                vendor = Vendor.STELLIO
+            else:
+                return None
+            return vendor, version
         except Exception:
+            if is_interactive():
+                print("Java-Spring based Context Broker detected. Try to enable info endpoint.")
             return None
 
-    def _broker_version_cassiopeia(self) -> Optional[str]:
-        """Requests the broker looking for Cassiopeia version.
-
-        Raises
-        ------
-        NotImplemented
-        """
-        raise NotImplementedError
-
     def _welcome_message(self) -> str:
-        return f"Connected to Context Broker at {self.hostname}:{self.port} | vendor={self.broker.vendor.value} version={self.broker.version}"
+        return f"Connected to Context Broker at {self.hostname}:{self.port} | vendor={self.broker.vendor.value} | version={self.broker.version}"
 
     def _fail_message(self) -> str:
         return f"Failed to connect to Context Broker at {self.hostname}:{self.port}"
 
+    def _warn_spring_message(self) -> str:
+        return "Java-Spring based Context Broker detected. Info endpoint disabled."
+
     # below the context manager methods
 
     def __enter__(self):
