@@ -20,12 +20,13 @@ from functools import partialmethod
 from dataclasses import dataclass
 
 from datetime import datetime
-from typing import overload, Any, Union, List, Optional
+from typing import overload, Any, Union, List, Optional, Callable
+from rich import print_json
 
 from .exceptions import *
 from .constants import *
 from .ngsidict import NgsiDict
-from ngsildclient.utils import iso8601, url
+from ngsildclient.utils import iso8601, url, is_interactive
 from ngsildclient.utils.urn import Urn
 
 logger = logging.getLogger(__name__)
@@ -162,6 +163,7 @@ class Entity:
 
         strict: bool = False  # for future use
         autoescape: bool = True  # for future use
+        f_print: Callable = print_json if is_interactive() else print
 
     globalsettings: Settings = Settings()
 
@@ -248,6 +250,9 @@ class Entity:
         if autoprefix is None:
             autoprefix = Entity.globalsettings.autoprefix
 
+        self._lastprop: NgsiDict = None
+        self._anchored: bool = False
+
         if payload is not None:  # create Entity from a dictionary
             if not payload.get("id", None):
                 raise NgsiMissingIdError()
@@ -286,8 +291,6 @@ class Entity:
             {"@context": ctx, "id": urn.fqn, "type": type},
             dtcached=dt if dt else iso8601.utcnow(),
         )
-        self._lastprop: NgsiDict = None
-        self._anchored: bool = False
 
     @classmethod
     def from_dict(cls, payload: dict):
@@ -308,41 +311,6 @@ class Entity:
         """
         return cls(payload=payload)
 
-    @classmethod
-    def load(cls, filename: str):
-        """Load an Entity from a JSON file, locally from the filesystem or remotely through HTTP.
-
-        For convenience some `SmartDataModels <https://smartdatamodels.org/>`_ examples are made available thanks to the Smart Data Models initiative.
-        You can benefit from autocompletion to navigate inside the available datamodels.
-
-        Parameters
-        ----------
-        filename : str
-            If filename corresponds to an URL, the JSON file is downloaded from HTTP.
-            Else it is retrieved locally from the filesystem.
-
-        Returns
-        -------
-        [Entity]
-            The Entity instance
-
-        See Also
-        --------
-        model.constants.SmartDataModels
-
-        Example:
-        --------
-        >>> from ngsildclient import *
-        >>> e = Entity.load(SmartDatamodels.SmartCities.Weather.WeatherObserved)
-        """
-        if url.isurl(filename):
-            resp = requests.get(filename)
-            d = resp.json()
-        else:
-            with open(filename, "r") as fp:
-                d = json.load(fp)
-        return cls.from_dict(d)
-
     @classmethod
     def duplicate(cls, entity: Entity) -> Entity:
         """Duplicate a given entity.
@@ -544,16 +512,21 @@ class Entity:
             }
         }
         """
-        property = self._payload.build_property(
-            value, unitcode, observedat, datasetid, userdata, escape
-        )
+        property = self._payload._build_property(value, unitcode, observedat, datasetid, userdata, escape)
         self._update_entity(name, property, nested)
         return self
 
     def addr(self, value: str):
         return self.prop("address", value)
 
-    def gprop(self, name: str, value: NgsiGeometry, nested: bool = False) -> Entity:
+    def gprop(
+        self,
+        name: str,
+        value: NgsiGeometry,
+        nested: bool = False,
+        observedat: Union[str, datetime] = None,
+        datasetid: str = None,
+    ) -> Entity:
         """Build a GeoProperty.
 
         Build a GeoProperty and attach it to the current entity.
@@ -600,7 +573,7 @@ class Entity:
             }
         }
         """
-        property = self._payload.build_geoproperty(value)
+        property = self._payload._build_geoproperty(value, observedat, datasetid)
         self._update_entity(name, property, nested)
         return self
 
@@ -652,7 +625,7 @@ class Entity:
         """
         if value is None:
             value = self._payload._dtcached
-        property = self._payload.build_temporal_property(value)
+        property = self._payload._build_temporal_property(value)
         self._update_entity(name, property, nested)
         return self
 
@@ -669,6 +642,7 @@ class Entity:
         nested: bool = False,
         *,
         observedat: Union[str, datetime] = None,
+        datasetid: str = None,
         userdata: NgsiDict = NgsiDict(),
     ) -> Entity:
         """Build a Relationship Property.
@@ -709,7 +683,7 @@ class Entity:
         if isinstance(name, Rel):
             name = name.value
 
-        property = self._payload.build_relationship(value, observedat, userdata)
+        property = self._payload._build_relationship(value, observedat, datasetid, userdata)
         self._update_entity(name, property, nested)
         return self
 
@@ -749,9 +723,7 @@ class Entity:
         d = NgsiDict()
         for k, v in self._payload.items():
             if isinstance(v, dict):
-                if (
-                    v["type"] == AttrType.PROP.value
-                ):  # apply to Property and TemporalProperty
+                if v["type"] == AttrType.PROP.value:  # apply to Property and TemporalProperty
                     value = v["value"]
                     if isinstance(value, dict):
                         value = value.get("@value", value)  # for Temporal Property only
@@ -788,7 +760,69 @@ class Entity:
         kv : bool, optional
             KeyValues format (aka simplified representation), by default False
         """
-        print(self.to_json(kv, indent=2, *args, **kwargs))
+        Entity.globalsettings.f_print(self.to_json(kv, indent=2, *args, **kwargs))
+
+    @classmethod
+    def load(cls, filename: str):
+        """Load an Entity from a JSON file, locally from the filesystem or remotely through HTTP.
+
+        For convenience some `SmartDataModels <https://smartdatamodels.org/>`_ examples are made available thanks to the Smart Data Models initiative.
+        You can benefit from autocompletion to navigate inside the available datamodels.
+
+        Parameters
+        ----------
+        filename : str
+            If filename corresponds to an URL, the JSON file is downloaded from HTTP.
+            Else it is retrieved locally from the filesystem.
+
+        Returns
+        -------
+        Entity
+            The Entity instance
+
+        See Also
+        --------
+        model.constants.SmartDataModels
+
+        Example:
+        --------
+        >>> from ngsildclient import *
+        >>> e = Entity.load(SmartDatamodels.SmartCities.Weather.WeatherObserved)
+        """
+        if url.isurl(filename):
+            resp = requests.get(filename)
+            payload = resp.json()
+        else:
+            with open(filename, "r") as fp:
+                payload = json.load(fp)
+        if isinstance(payload, List):
+            return [cls.from_dict(x) for x in payload]
+        return cls.from_dict(payload)
+
+    @classmethod
+    def load_batch(cls, filename: str):
+        """Load a batch of entities from a JSON file.
+
+        Parameters
+        ----------
+        filename : str
+            The input file must contain a JSON array
+
+        Returns
+        -------
+        List[Entity]
+            A list of entities
+
+        Example:
+        --------
+        >>> from ngsildclient import *
+        >>> rooms = Entity.load_batch("/tmp/rooms_all.jsonld")
+        """
+        with open(filename, "r") as fp:
+            payload = json.load(fp)
+        if not isinstance(payload, List):
+            raise ValueError("The JSON payload MUST be an array")
+        return [cls.from_dict(x) for x in payload]
 
     def save(self, filename: str, *, indent: int = 2):
         """Save the entity to a file.
@@ -802,3 +836,26 @@ class Entity:
         """
         with open(filename, "w") as fp:
             json.dump(self._payload, fp, default=str, ensure_ascii=False, indent=indent)
+
+    @classmethod
+    def save_batch(cls, entities: List[Entity], filename: str, *, indent: int = 2):
+        """Save a batch of entities to a JSON file.
+
+        Parameters
+        ----------
+        entities: List[Entity]
+            Batch of entities to be saved
+
+        filename : str
+            If filename corresponds to an URL, the JSON file is downloaded from HTTP.
+            Else it is retrieved locally from the filesystem.
+
+        Example:
+        --------
+        >>> from ngsildclient import *
+        >>> rooms = [Entity("Room", "Room1"), Entity("Room", "Room2")]
+        >>> Entity.save_batch(rooms, "/tmp/rooms_all.jsonld")
+        """
+        payload = [x._payload for x in entities]
+        with open(filename, "w") as fp:
+            json.dump(payload, fp, default=str, ensure_ascii=False, indent=indent)
